//PROFILE-NO
unit ScanFrost;

// *****************************************************************************
// * Copyright 2003 mxbee                                                      *
// *****************************************************************************
// * This program is free software; you can redistribute it and/or modify      *
// * it under the terms of the GNU General Public License as published by      *
// * the Free Software Foundation; either version 2 of the License, or         *
// * (at your option) any later version.                                       *
// *                                                                           *
// * This program is distributed in the hope that it will be useful,           *
// * but WITHOUT ANY WARRANTY; without even the implied warranty of            *
// * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
// * GNU General Public License for more details.                              *
// *                                                                           *
// * You should have received a copy of the GNU General Public License         *
// * along with this program; if not, write to the Free Software               *
// * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
// *****************************************************************************

{$INCLUDE CompilerOpts.pas}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, CheckLst, Mask, Menus;

type
  TFrmScanFrost = class(TForm)
    Panel1: TPanel;
    LbFrostFolder: TLabel;
    BtFrostFolder: TButton;
    Label1: TLabel;
    PanelBoards: TPanel;
    Label2: TLabel;
    ClbBoards: TCheckListBox;
    OpenDlgFrostFolder: TOpenDialog;
    Panel2: TPanel;
    Label3: TLabel;
    EdMinDate: TEdit;
    Label5: TLabel;
    EdMaxDate: TEdit;
    BtOk: TButton;
    PopupMenu1: TPopupMenu;
    MiCheckAll: TMenuItem;
    MiUncheckAll: TMenuItem;
    procedure BtFrostFolderClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure BtOkClick(Sender: TObject);
    procedure MiCheckAllClick(Sender: TObject);
  private
    FFrostDir:   String;
    FFileList:   String;
    procedure UpdateBoardList;
    procedure SetAnchors;
  public
    property Files: String read FFileList;
  end;

var
  FrmScanFrost: TFrmScanFrost = nil;

implementation

// avoid FileCtrl is platform specific warning
{$IFDEF DELPHI_6_OR_HIGHER} {$WARN UNIT_PLATFORM OFF} {$ENDIF}
uses FileCtrl;

{$R *.dfm}

procedure TFrmScanFrost.FormCreate(Sender: TObject);
begin
  SetAnchors;
end;

procedure TFrmScanFrost.BtFrostFolderClick(Sender: TObject);
var s: String;
begin
  if not OpenDlgFrostFolder.Execute then exit;
  s := ExtractFilePath(OpenDlgFrostFolder.Filename);
  if not DirectoryExists(s + 'keypool') then begin
    MessageDlg('Folder ' + s + 'keypool does not exists.', mtError, [mbOk], 0);
    exit;
  end;
  FFrostDir := s;
  LbFrostFolder.Caption := FFrostDir;
  UpdateBoardList;
end;

procedure TFrmScanFrost.UpdateBoardList;
var
  SearchRec: TSearchRec;
begin
  ClbBoards.Items.Clear;
  if 0 = FindFirst(FFrostDir + 'keypool\*.*', faDirectory, SearchRec) then begin
    ClbBoards.Items.BeginUpdate;
    try
      repeat
        if (SearchRec.Name = '.') or (SearchRec.Name = '..') then continue;
        ClbBoards.Items.Add(SearchRec.Name);
      until 0 <> FindNext(SearchRec);
    finally
      ClbBoards.Items.EndUpdate;
      FindClose(SearchRec);
    end;
  end;
end;

procedure TFrmScanFrost.BtOkClick(Sender: TObject);

  function  MyStrToDate(Str: String): TDateTime;
  begin
    Str := Trim(Str);
    if Str = '' then
      Result := 0
    else begin
      try
        if Length(Str) <> 10 then raise EAbort.Create('');
        if (Str[5] <> '-') or (Str[8] <> '-') then raise EAbort.Create('');
        Result := EncodeDate(StrToInt(Copy(Str,1,4)),StrToInt(Copy(Str,6,2)),StrToInt(Copy(Str,9,2)));
      except
        raise Exception.Create('Invalid date "' + Str + '" (use format YYYY-MM-DD)!');
      end;
    end;
  end;

var
  dtMin,dtMax,dt: TDateTime;
  SearchRec,SRec2: TSearchRec;
  slFiles:     TStringList;
  iBoard,i:    Integer;
  s,sBoardDir: String;
  y,m,d:       Word;
begin
  dtMin := MyStrToDate(EdMinDate.Text);
  dtMax := MyStrToDate(EdMaxDate.Text);

  slFiles := nil;
  try
    Screen.Cursor := crHourglass;
    slFiles := TStringList.Create;

    for iBoard := 0 to ClbBoards.Items.Count-1 do begin
      if not ClbBoards.Checked[iBoard] then continue;
      sBoardDir := FFrostDir + 'keypool\' + ClbBoards.Items[iBoard] + '\';
      if 0 = FindFirst(sBoardDir + '*.*', faDirectory, SearchRec) then begin
        try
          repeat
            if (SearchRec.Name = '.') or (SearchRec.Name = '..') then continue;

            // parse dir name (frost uses year.month.day)
            s := SearchRec.Name;
            i := Pos('.', s); if i = 0 then continue;
            y := StrToIntDef(Copy(s,1,i-1), 0); Delete(s,1,i);
            i := Pos('.', s); if i = 0 then continue;
            m := StrToIntDef(Copy(s,1,i-1), 0); Delete(s,1,i);
            d := StrToIntDef(s, 0);
            if (y = 0) or (m < 1) or (m > 12) or (d < 1) or (d > 31) then continue;
            try
              dt := EncodeDate(y,m,d);
              if (dtMin <> 0) and (dt < dtMin) then continue;
              if (dtMax <> 0) and (dt > dtMax) then continue;
            except
              continue;
            end;

            // ok, get the xml files for that day
            if 0 = FindFirst(sBoardDir + SearchRec.Name + '\*.xml', 0, SRec2) then begin
              try
                repeat
                  slFiles.Add(sBoardDir + SearchRec.Name + '\' + SRec2.Name);
                until 0 <> FindNext(SRec2);
              finally
                FindClose(SRec2);
              end;
            end;

          until 0 <> FindNext(SearchRec);
        finally
          FindClose(SearchRec);
        end;
      end;
    end;

    FFileList := slFiles.Text;

  finally
    Screen.Cursor := crDefault;
    slFiles.Free;
  end;

  ModalResult := mrOk;
end;

procedure TFrmScanFrost.MiCheckAllClick(Sender: TObject);
var
  bCheck: Boolean;
  i:      Integer;
begin
  bCheck := (Sender = MiCheckAll);
  ClbBoards.Items.BeginUpdate;
  try
    for i := 0 to ClbBoards.Items.Count-1 do
      ClbBoards.Checked[i] := bCheck;
  finally
    ClbBoards.Items.EndUpdate;
  end;
end;

procedure TFrmScanFrost.SetAnchors;
begin
  // Set anchors for controls that stick to the right or bottom of their parent.
  // Normally this would be done at design time, but D7 then saves .DesignSize,
  // breaking compatibility with earlier Delphi versions. So do it at runtime.
  with LbFrostFolder          do Anchors := [akLeft,akTop,akRight];
  with BtFrostFolder          do Anchors := [akTop,akRight];
  with ClbBoards              do Anchors := [akLeft,akTop,akRight,akBottom];
end;

end.
